# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import logging
from typing import List, Optional

# project
from kiwi.utils.temporary import Temporary
from kiwi.command import Command
from kiwi.partitioner.base import PartitionerBase
from kiwi.utils.block import BlockID

from kiwi.exceptions import (
    KiwiPartitionerMsDosFlagError
)

log = logging.getLogger('kiwi')


class PartitionerMsDos(PartitionerBase):
    """
    **Implement old style msdos partition setup**
    """
    def post_init(self) -> None:
        """
        Post initialization method

        Setup sfdisk partition type/flag map
        """
        self.default_start = 2048
        self.sector_size = 512
        self.flag_map = {
            'f.active': True,
            't.linux': '83',
            't.swap': '82',
            't.lvm': '8e',
            't.raid': 'fd',
            't.efi': None,
            't.csm': None,
            't.prep': '41',
            't.extended': '5'
        }

    def create(
        self, name: str, mbsize: int, type_name: str, flags: List[str] = [],
        partition_id: Optional[int] = None
    ) -> None:
        """
        Create msdos partition

        :param string name: partition name
        :param int mbsize: partition size
        :param string type_name: partition type
        :param list flags: additional flags
        :param int partition_id:
            If provided, use this exact partition ID instead
            of auto-incrementing.
        """
        if self.extended_layout:
            if self.partition_count < 3:
                # in primary boundary
                self._create_primary(
                    name, mbsize, type_name, flags, partition_id
                )
            elif self.partition_count == 3:
                # at primary boundary, create extended + logical
                self._create_extended(name)
                self._create_logical(
                    name, mbsize, type_name, flags, partition_id
                )
            elif self.partition_count > 3:
                # in logical boundary
                self._create_logical(
                    name, mbsize, type_name, flags, partition_id
                )
        else:
            self._create_primary(
                name, mbsize, type_name, flags, partition_id
            )

    def set_flag(self, partition_id: int, flag_name: str) -> None:
        """
        Set msdos partition flag

        :param int partition_id: partition number
        :param string flag_name: name from flag map
        """
        if flag_name not in self.flag_map:
            raise KiwiPartitionerMsDosFlagError(
                f'Unknown partition flag {flag_name}'
            )
        flag_val = self.flag_map[flag_name]
        if flag_val:
            if flag_name == 'f.active':
                Command.run(
                    [
                        'parted', self.disk_device,
                        'set', format(partition_id), 'boot', 'on'
                    ]
                )
            else:
                assert isinstance(flag_val, str), f"flag {flag_name} must be a string but got a {type(flag_val)}"
                Command.run(
                    [
                        'sfdisk', '-c', self.disk_device,
                        format(partition_id), flag_val
                    ]
                )
        else:
            log.warning('Flag %s ignored on msdos', flag_name)

    def resize_table(self, entries: int = None) -> None:
        """
        Resize partition table

        Nothing to be done here for msdos table

        :param int entries: unused
        """
        pass

    def set_uuid(self, partition_id: int, uuid: str) -> None:
        """
        Set partition UUID

        Nothing to be done here for MSDOS devices

        :param int partition_id: unused
        :param string uuid: unused
        """
        pass  # pragma: nocover

    def set_start_sector(self, start_sector: int):
        """
        Set start sector of first partition as configured.
        fdisk and friends are not able to work correctly if
        the start sector of the first partition is any different
        from 2048.

        :param int start_sector: sector size
        """
        block_operation = BlockID(self.disk_device)
        fdisk_input = Temporary().new_file()
        with open(fdisk_input.name, 'w') as partition:
            if block_operation.get_partition_count() >= 4:
                # if the partition table is full fdisk does not ask
                # for the partition number when there is only one choice
                log.debug(f'fdisk: d 1 n p {start_sector} w q')
                partition.write(
                    'd\n1\nn\np\n{0}\n\nw\nq\n'.format(start_sector)
                )
            else:
                # if the partition table has less than 4 partitions
                # fdisk will ask for the partition number to change
                log.debug(f'fdisk: d 1 n p 1 {start_sector} w q')
                partition.write(
                    'd\n1\nn\np\n1\n{0}\n\nw\nq\n'.format(start_sector)
                )
        self._call_fdisk(fdisk_input.name)

    def _create_primary(
        self, name: str, mbsize: int, type_name: str, flags: List[str] = [],
        partition_id: Optional[int] = None
    ) -> None:
        """
        Create primary msdos partition
        """
        self.partition_id = \
            partition_id if partition_id else self.partition_id + 1
        self.partition_count += 1
        fdisk_input = Temporary().new_file()
        if self.partition_id == 1 and self.start_sector:
            if self.start_sector > self.default_start and mbsize != 'all_free':
                # fdisk default start sector is self.default_start, increase
                # the partition size such that after set_start_sector()
                # the requested partition size is present
                mbsize += int(
                    (self.start_sector - self.default_start) * self.sector_size / (1024 * 1024)
                )
        with open(fdisk_input.name, 'w') as partition:
            log.debug(
                '%s: fdisk: n p %d cur_position +%sM w q',
                name, self.partition_id, format(mbsize)
            )
            partition.write(
                'n\np\n{0}\n\n{1}\nw\nq\n'.format(
                    self.partition_id,
                    '' if mbsize == 'all_free' else '+{0}M'.format(mbsize)
                )
            )
        self._call_fdisk(fdisk_input.name)
        self._set_all_flags(type_name, flags)

    def _create_extended(
        self, name: str, partition_id: Optional[int] = None
    ) -> None:
        """
        Create extended msdos partition
        """
        self.partition_id = \
            partition_id if partition_id else self.partition_id + 1
        self.partition_count += 1
        fdisk_input = Temporary().new_file()
        with open(fdisk_input.name, 'w') as partition:
            log.debug(
                '%s: fdisk: n e %d cur_position +all_freeM w q',
                name, self.partition_id
            )
            partition.write(
                'n\ne\n{0}\n{1}\n{2}\nw\nq\n'.format(
                    self.partition_id, '', ''
                )
            )
        self._call_fdisk(fdisk_input.name)

    def _create_logical(
        self, name: str, mbsize: int, type_name: str, flags: List[str] = [],
        partition_id: Optional[int] = None
    ) -> None:
        """
        Create logical msdos partition
        """
        self.partition_id = \
            partition_id if partition_id else self.partition_id + 1
        self.partition_count += 1
        fdisk_input = Temporary().new_file()
        with open(fdisk_input.name, 'w') as partition:
            log.debug(
                '%s: fdisk: n %d cur_position +%sM w q',
                name, self.partition_id, format(mbsize)
            )
            partition.write(
                'n\n{0}\n{1}\n{2}\nw\nq\n'.format(
                    self.partition_id,
                    '',
                    '' if mbsize == 'all_free' else '+{0}M'.format(mbsize)
                )
            )
        self._call_fdisk(fdisk_input.name)
        self._set_all_flags(type_name, flags)

    def _set_all_flags(self, type_name: str, flags: List[str]) -> None:
        self.set_flag(self.partition_id, type_name)
        if flags:
            for flag_name in flags:
                self.set_flag(self.partition_id, flag_name)

    def _call_fdisk(self, fdisk_config_file_path: str) -> None:
        bash_command = ' '.join(
            ['cat', fdisk_config_file_path, '|', 'fdisk', self.disk_device]
        )
        try:
            Command.run(
                ['bash', '-c', bash_command]
            )
        except Exception:
            # unfortunately fdisk reports that it can't read in the partition
            # table which I consider a bug in fdisk. However the table was
            # correctly created and therefore we continue. Problem is that we
            # are not able to detect real errors with the fdisk operation at
            # that point.
            log.debug('potential fdisk errors were ignored')



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os

# project
from kiwi.command import Command
from kiwi.defaults import Defaults
from kiwi.utils.command_capabilities import CommandCapabilities


class ArchiveTar:
    """
    **Extraction/Creation of tar archives**

    The tarfile python module is not used by that class,
    since it does not provide support for some relevant features
    in comparison to the GNU tar command (e.g. numeric-owner).
    Moreover tarfile lacks support for xz compression under
    Python v2.7.

    :param string filename:
        filename to use for archive extraction or creation
    :param bool create_from_file_list:
        use file list not entire directory to create the archive
    :param list file_list:
        list of files and directorie names to archive
    """
    def __init__(self, filename, create_from_file_list=True, file_list=None):
        self.filename = filename
        self.create_from_file_list = create_from_file_list
        self.file_list = file_list

        if CommandCapabilities.check_version('tar', (1, 27)):
            self.xattrs_options = [
                '--xattrs', '--xattrs-include=*'
            ]
        else:
            self.xattrs_options = []

    def create(self, source_dir, exclude=None, options=None):
        """
        Create uncompressed tar archive

        :param string source_dir: data source directory
        :param list exclude: list of excluded items
        :param list options: custom creation options
        """
        if not options:
            options = []
        Command.run(
            [
                'tar', '-C', source_dir
            ] + options + self.xattrs_options + [
                '-c', '-f', self.filename
            ] + self._get_archive_items(source_dir, exclude)
        )
        return self.filename

    def append_files(self, source_dir, files_to_append, options=None):
        """
        Append files to an already existing uncompressed tar archive

        :param string source_dir: data source directory
        :param list files_to_append: list of items to append
        :param list options: custom options
        """
        if not options:
            options = []
        Command.run(
            [
                'tar', '-C', source_dir, '-r',
                '--file=' + self.filename
            ] + options + self.xattrs_options + files_to_append
        )
        return self.filename

    def create_xz_compressed(
        self, source_dir, exclude=None, options=None, xz_options=None
    ):
        """
        Create XZ compressed tar archive

        :param string source_dir: data source directory
        :param list exclude: list of excluded items
        :param list options: custom tar creation options
        :param list xz_options: custom xz compression options
        """
        if not options:
            options = []
        if not xz_options:
            xz_options = Defaults.get_xz_compression_options()
        bash_command = [
            'tar', '-C', source_dir
        ] + options + self.xattrs_options + [
            '-c', '--to-stdout'
        ] + self._get_archive_items(source_dir, exclude) + [
            '|', 'xz', '-f'
        ] + xz_options + [
            '>', self.filename + '.xz'
        ]
        Command.run(['bash', '-c', ' '.join(bash_command)])
        return self.filename + '.xz'

    def create_gnu_gzip_compressed(self, source_dir, exclude=None):
        """
        Create gzip compressed tar archive

        :param string source_dir: data source directory
        :param list exclude: list of excluded items
        """
        Command.run(
            [
                'tar', '-C', source_dir,
                '--format=gnu', '-cSz', '-f', self.filename + '.gz'
            ] + self._get_archive_items(source_dir, exclude)
        )
        return self.filename + '.gz'

    def extract(self, dest_dir):
        """
        Extract tar archive contents

        :param string dest_dir: target data directory
        """
        Command.run(
            ['tar', '-C', dest_dir, '-x', '-v', '-f', self.filename]
        )

    def _get_archive_items(self, source_dir, exclude_list):
        if self.create_from_file_list:
            return self._get_archive_content_list(
                source_dir, exclude_list, self.file_list
            )
        else:
            return self._get_archive_exclude_list(exclude_list) + ['.']

    def _get_archive_content_list(
        self, source_dir, exclude_list=None, file_list=None
    ):
        final_archive_contents = []
        archive_contents = file_list
        if not archive_contents:
            archive_contents = sorted(os.listdir(source_dir))

        for item in archive_contents:
            if not exclude_list:
                final_archive_contents.append(item)
            elif item not in exclude_list:
                final_archive_contents.append(item)

        return final_archive_contents

    def _get_archive_exclude_list(self, exclude_list):
        archive_items = []
        for exclude in exclude_list:
            archive_items.append('--exclude')
            archive_items.append('./' + exclude)
        return archive_items



# Copyright (c) 2024 SUSE LLC.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
from typing import List
from string import Template
from textwrap import dedent


class BuilderTemplateSystemdUnit:
    """
    **systemd unit file templates**
    """
    def __init__(self):
        self.unit = dedent('''
            # kiwi generated unit file
            [Unit]
            Description=Import Local Container(s)
        ''').strip() + os.linesep

        self.service = dedent('''
            [Service]
            Type=oneshot
        ''').strip() + os.linesep

        self.install = dedent('''
            [Install]
            WantedBy=multi-user.target
        ''').strip() + os.linesep

    def get_container_import_template(
        self, container_files: List[str], load_commands: List[List[str]],
        after: List[str]
    ):
        template_data = self.unit
        for container_file in container_files:
            template_data += 'ConditionPathExists={0}{1}'.format(
                container_file, os.linesep
            )
        if after:
            template_data += 'After={0}{1}'.format(
                ' '.join(after), os.linesep
            )
        template_data += self.service
        for load_command in load_commands:
            template_data += 'ExecStart={0}{1}'.format(
                ' '.join(load_command), os.linesep
            )
        for container_file in container_files:
            template_data += 'ExecStartPost=/bin/rm -f {0}{1}'.format(
                container_file, os.linesep
            )
        template_data += self.install
        return Template(template_data)



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import importlib
from typing import Dict
from abc import (
    ABCMeta,
    abstractmethod
)

# project
from kiwi.exceptions import KiwiSnapshotManagerSetupError


class SnapshotManager(metaclass=ABCMeta):
    """
    **SnapshotManager factory**

    :param str name: snapshot management name
    :param str device: storage device node name
    :param str root_dir: root directory path
    :param str mountpoint: mountpoint of the filesystem to snapshot
    :param str root_volume_name: the name of the root volume in case
        snapshots are hosted in a subvolume.
    :param dict custom_args: dictionary of custom snapshot manager arguments
    """
    @abstractmethod
    def __init__(self) -> None:
        return None  # pragma: no cover

    @staticmethod
    def new(
        name: str, device: str, root_dir: str, mountpoint: str,
        root_volume_name: str, custom_args: Dict = None
    ):
        name_map = {
            'snapper': 'Snapper',
        }
        try:
            snapshot_manager = importlib.import_module(
                'kiwi.snapshot_manager.{0}'.format(name)
            )
            module_name = 'SnapshotManager{0}'.format(name_map[name])
            return snapshot_manager.__dict__[module_name](
                device, root_dir, mountpoint, root_volume_name, custom_args
            )
        except Exception as issue:
            raise KiwiSnapshotManagerSetupError(
                'Support for {0} snapshot manager not implemented: {1}'.format(
                    name, issue
                )
            )



# Copyright (c) 2018 Eaton.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import pathlib
from textwrap import dedent

# project
from kiwi.firmware import FirmWare
from kiwi.storage.subformat.vmdk import DiskFormatVmdk
from kiwi.storage.subformat.base import DiskFormatBase
from kiwi.command import Command
from kiwi.path import Path
from kiwi.system.result import Result
from kiwi.storage.subformat.template.ova_compose import (
    OvaComposeTemplate
)

from kiwi.exceptions import (
    KiwiFormatSetupError,
    KiwiCommandNotFound,
    KiwiTemplateError
)


class DiskFormatOva(DiskFormatBase):
    """
    **Create ova disk format, based on vmdk**
    """
    def post_init(self, custom_args: dict) -> None:
        """
        vmdk disk format post initialization method

        Store qemu options as list from custom args dict

        :param dict custom_args: custom qemu arguments dictionary
        """
        ovftype = 'vmware'  # default OVA target infrastructure
        self.machine_setup = self.xml_state.get_build_type_machine_section()
        if self.machine_setup and self.machine_setup.get_ovftype():
            ovftype = self.machine_setup.get_ovftype()
        if ovftype != 'vmware':
            raise KiwiFormatSetupError(f'Unsupported ovftype {ovftype}')
        self.image_format = 'ova'
        # ensure streamOptimized format is enabled
        custom_args.update(
            {'subformat=streamOptimized': None}
        )
        self.firmware = FirmWare(
            self.xml_state
        )
        self.vmdk = DiskFormatVmdk(
            self.xml_state, self.root_dir, self.target_dir, custom_args
        )

    def create_image_format(self) -> None:
        """
        Create ova disk format using ova-compose
        https://github.com/vmware/open-vmdk
        """
        # Check for required mkova tool
        ova_compose = Path.which(filename='ova-compose', access_mode=os.X_OK)
        if not ova_compose:
            tool_not_found_message = dedent('''\n
                Required ova-compose not found in PATH on the build host

                Building OVA images requires the ova-compose tool which
                is usually provided by the open-vmdk package.
            ''')
            raise KiwiCommandNotFound(
                tool_not_found_message
            )

        # Create the vmdk disk image and vmx config
        self.vmdk.create_image_format()

        compose_meta = self.get_target_file_path_for_format('meta')
        ova = self.get_target_file_path_for_format('ova')

        # Create ova composer meta file
        self._create_ova_compose_meta_file()

        pathlib.Path(ova).unlink(missing_ok=True)
        os.chdir(self.target_dir)
        Command.run(
            [
                ova_compose,
                '--input-file', compose_meta,
                '--output-file', ova,
                '--format', 'ova'
            ]
        )

    def store_to_result(self, result: Result) -> None:
        """
        Store the resulting ova file into the provided result instance.

        :param object result: Instance of Result
        """
        result.add(
            key='disk_format_image',
            filename=self.get_target_file_path_for_format('ova'),
            use_for_bundle=True,
            compress=self.runtime_config.get_bundle_compression(
                default=False
            ),
            shasum=True
        )

    def _create_ova_compose_meta_file(self) -> None:
        """
        Create the ova-compose meta file used to create the final ova file
        """
        displayname = self.xml_state.xml_data.get_displayname()
        efi_mode = self.firmware.efi_mode()
        template_record = {
            'display_name':
                displayname or self.xml_state.xml_data.get_name(),
            'vmdk_file':
                os.path.basename(self.get_target_file_path_for_format('vmdk')),
            'guest_os': 'other4xLinux64Guest',
            'memory_size': 4096,
            'number_of_cpus': 2,
            'firmware': 'efi' if efi_mode else 'bios',
            'secure_boot': 'true' if efi_mode == 'uefi' else 'false'
        }

        # Basic setup
        machine_setup = self.xml_state.get_build_type_machine_section()
        if machine_setup:
            memory_setup = machine_setup.get_memory()
            guest_os = machine_setup.get_guestOS()
            cpu_setup = machine_setup.get_ncpus()
            if guest_os:
                template_record['guest_os'] = guest_os
            if memory_setup:
                template_record['memory_size'] = memory_setup
            if cpu_setup:
                template_record['number_of_cpus'] = cpu_setup

        # Build settings template and write settings file
        settings_template = OvaComposeTemplate().get_template(
            bool(machine_setup)
        )
        try:
            settings_file = self.get_target_file_path_for_format('meta')
            with open(settings_file, 'w') as config:
                config.write(settings_template.substitute(template_record))
        except Exception as issue:
            raise KiwiTemplateError(
                f'{type(issue).__name__}: {format(issue)}'
            )



# Copyright (c) 2026 Marcus Schäfer.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
from pathlib import Path

# project
from kiwi.storage.subformat.base import DiskFormatBase
from kiwi.utils.temporary import Temporary
from kiwi.system.root_init import RootInit
from kiwi.system.root_import import RootImport
from kiwi.system.result import Result
from kiwi.command import Command
from kiwi.storage.subformat import DiskFormat
from kiwi.defaults import Defaults
from kiwi.runtime_config import RuntimeConfig
from kiwi.exceptions import KiwiContainerBuilderError
from kiwi.utils.checksum import Checksum
from kiwi.container import ContainerImage


class DiskFormatOci(DiskFormatBase):
    """
    **Create OCI container which includes the disk image**
    """
    def post_init(self, custom_args: dict) -> None:
        """
        oci disk format post initialization method

        :param dict custom_args: unused
        """
        self.runtime_config = RuntimeConfig()
        self.ensure_empty_tmpdirs = True
        self.image_format: str = 'oci'
        self.container_config = self.xml_state.get_container_config()
        self.bundle_format = self.xml_state.get_build_type_bundle_format()
        self.base_container_uris = self.xml_state.get_derived_from_image_uri()
        self.base_format = custom_args.get(
            'base_format', 'raw'
        )
        self.requested_container_type = custom_args.get(
            'container_format', 'docker'
        )
        self.special_needs = ['appx', 'wsl']
        self.filename = ''.join(
            [
                self.target_dir, '/',
                self.xml_state.xml_data.get_name(),
                '.' + Defaults.get_platform_name(),
                '-' + self.xml_state.get_image_version(),
                '.', self.requested_container_type,
                '.tar' if self.requested_container_type not in
                self.special_needs else ''
            ]
        )

    def create_image_format(self) -> None:
        """
        Create oci container and store disk to it
        """
        base_image = None
        container_rootdir = Temporary(prefix='kiwi_container_root.').new_dir()
        container_root = container_rootdir.name
        if self.base_container_uris:
            root = RootInit(
                container_root, allow_existing=True
            )
            root.create()
            root_import = RootImport.new(
                container_root,
                self.base_container_uris,
                self.requested_container_type
            )
            root_import.sync_data()
            base_image = Defaults.get_imported_root_image(
                container_root
            )
            base_image_sha256 = ''.join([base_image, '.sha256'])

        # Create disk format and copy this format or the raw
        # disk to the imported container root dir.
        disk_target = os.sep.join([container_root, 'disk'])
        Path(disk_target).mkdir(parents=True, exist_ok=True)

        if self.bundle_format:
            result = Result(self.xml_state)
            result.add_bundle_format(
                self.xml_state.get_build_type_bundle_format()
            )
            if result.get_name_for_pattern():
                disk_target = '{}/{}.{}'.format(
                    disk_target,
                    result.get_name_for_pattern(),
                    self.base_format
                )

        if self.base_format != 'raw':
            with DiskFormat.new(
                self.base_format, self.xml_state,
                self.root_dir, self.target_dir
            ) as disk_format:
                disk_format.create_image_format()
                disk_format_file = disk_format.get_target_file_path_for_format(
                    self.base_format
                )
                disk_raw_file_stat = os.stat(self.diskname)
                os.chown(
                    disk_format_file,
                    disk_raw_file_stat.st_uid,
                    disk_raw_file_stat.st_gid
                )
                Command.run(
                    ['cp', '-a', disk_format_file, disk_target]
                )
        else:
            Command.run(
                ['cp', '-a', self.diskname, disk_target]
            )

        # build the container
        if base_image:
            checksum = Checksum(base_image)
            if not checksum.matches(checksum.sha256(), base_image_sha256):
                raise KiwiContainerBuilderError(
                    f'base image file {base_image} checksum validation failed'
                )
        container_image = ContainerImage.new(
            self.requested_container_type,
            container_root,
            self.container_config
        )
        self.filename = container_image.create(
            self.filename, base_image or '', self.ensure_empty_tmpdirs,
            self.runtime_config.get_container_compression()
            # appx containers already contains a compressed root
            # wsl containers recommends gzip and we default to it
            if self.requested_container_type not in
            self.special_needs else False
        )

    def store_to_result(self, result: Result) -> None:
        """
        Store result file of the format conversion into the
        provided result instance.

        In case of a oci format we store the result uncompressed
        Since the created container is already compressed

        :param object result: Instance of Result
        """
        result.add(
            key='disk_format_image',
            filename=self.filename,
            use_for_bundle=True,
            compress=False,
            shasum=True
        )



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import logging
from typing import (
    Dict, Optional
)
import shutil

# project
from kiwi.utils.temporary import Temporary
from kiwi.command import Command
from kiwi.storage.device_provider import DeviceProvider
from kiwi.boot.image.base import BootImageBase
from kiwi.boot.image import BootImage
from kiwi.bootloader.config import create_boot_loader_config
from kiwi.filesystem.squashfs import FileSystemSquashFs
from kiwi.filesystem.isofs import FileSystemIsoFs
from kiwi.firmware import FirmWare
from kiwi.system.identifier import SystemIdentifier
from kiwi.path import Path
from kiwi.utils.block import BlockID
from kiwi.defaults import Defaults
from kiwi.utils.checksum import Checksum
from kiwi.system.kernel import Kernel
from kiwi.utils.compress import Compress
from kiwi.archive.tar import ArchiveTar
from kiwi.system.setup import SystemSetup
from kiwi.iso_tools.base import IsoToolsBase
from kiwi.xml_state import XMLState


from kiwi.exceptions import (
    KiwiInstallBootImageError
)

log = logging.getLogger('kiwi')


class InstallImageBuilder:
    """
    **Installation image builder**

    :param object xml_state: instance of :class:`XMLState`
    :param str root_dir: system image root directory
    :param str target_dir: target directory path name
    :param object boot_image_task: instance of :class:`BootImage`
    :param dict custom_args: Custom processing arguments defined as hash keys:
        * xz_options: string of XZ compression parameters
    """
    def __init__(
        self, xml_state: XMLState, root_dir: str, target_dir: str,
        boot_image_task: Optional[BootImageBase],
        custom_args: Dict = None
    ) -> None:
        self.arch = Defaults.get_platform_name()
        self.bootloader = xml_state.get_build_type_bootloader_name()
        if self.bootloader != 'systemd_boot':
            self.bootloader = 'grub2'
        self.root_dir = root_dir
        self.target_dir = target_dir
        self.xml_state = xml_state
        self.root_filesystem_is_multipath = \
            xml_state.get_oemconfig_oem_multipath_scan()
        self.initrd_system = xml_state.get_initrd_system()
        self.firmware = FirmWare(xml_state)
        self.setup = SystemSetup(
            self.xml_state, self.root_dir
        )
        self.iso_volume_id = self.xml_state.build_type.get_volid() or \
            Defaults.get_install_volume_id()
        self.diskname = ''.join(
            [
                target_dir, '/',
                xml_state.xml_data.get_name(),
                '.' + self.arch,
                '-' + xml_state.get_image_version(),
                '.raw'
            ]
        )
        self.isoname = ''.join(
            [
                target_dir, '/',
                xml_state.xml_data.get_name(),
                '.' + self.arch,
                '-' + xml_state.get_image_version(),
                '.install.iso'
            ]
        )
        self.pxename = ''.join(
            [
                xml_state.xml_data.get_name(),
                '.' + self.arch,
                '-' + xml_state.get_image_version()
            ]
        )
        self.pxetarball = ''.join(
            [
                target_dir, '/', self.pxename, '.install.tar'
            ]
        )
        self.dracut_config_file = ''.join(
            [self.root_dir, Defaults.get_dracut_conf_name()]
        )
        self.squashed_diskname = ''.join(
            [xml_state.xml_data.get_name(), '.raw']
        )
        self.sha256name = ''.join(
            [xml_state.xml_data.get_name(), '.sha256']
        )
        self.xz_options = custom_args['xz_options'] if custom_args \
            and 'xz_options' in custom_args else None

        self.mbrid = SystemIdentifier()
        self.mbrid.calculate_id()

        self.custom_iso_args: Dict = {}

        if not boot_image_task:
            self.boot_image_task = BootImage.new(
                xml_state, target_dir, root_dir
            )
            self.boot_image_task.prepare()
        else:
            self.boot_image_task = boot_image_task

    def create_install_iso(self) -> None:
        """
        Create an install ISO from the disk_image as hybrid ISO
        bootable via legacy BIOS, EFI and as disk from Stick

        Image types which triggers this builder are:

        * installiso="true|false"
        * installstick="true|false"
        """
        self.media_dir = Temporary(
            prefix='kiwi_install_media.', path=self.target_dir
        ).new_dir()
        # unpack cdroot user files to media dir
        self.setup.import_cdroot_files(self.media_dir.name)

        # custom iso metadata
        self.custom_iso_args = {
            'meta_data': {
                'volume_id': self.iso_volume_id,
                'mbr_id': self.mbrid.get_id(),
                'application_id': self.xml_state.build_type.get_application_id(),
                'efi_mode': self.firmware.efi_mode(),
                'efi_partition_table': self.firmware.get_partition_table_type(),
                'gpt_hybrid_mbr': self.firmware.gpt_hybrid_mbr,
                'ofw_mode': self.firmware.ofw_mode(),
                'legacy_bios_mode': self.firmware.legacy_bios_mode()
            }
        }

        # the system image transfer is checked against a checksum
        log.info('Creating disk image checksum')
        self.squashed_contents = Temporary(
            prefix='kiwi_install_squashfs.', path=self.target_dir
        ).new_dir()
        checksum = Checksum(self.diskname)
        checksum.sha256(self.squashed_contents.name + '/' + self.sha256name)

        # the system image name is stored in a config file
        self._write_install_image_info_to_iso_image()
        if self.initrd_system == 'kiwi':
            self._write_install_image_info_to_boot_image()

        # the system image is stored as squashfs embedded file
        log.info('Creating squashfs embedded disk image')
        Command.run(
            [
                'cp', '-l', self.diskname,
                self.squashed_contents.name + '/' + self.squashed_diskname
            ]
        )
        squashed_image_file = ''.join(
            [
                self.target_dir, '/', self.squashed_diskname, '.squashfs'
            ]
        )
        with FileSystemSquashFs(
            device_provider=DeviceProvider(),
            root_dir=self.squashed_contents.name,
            custom_args={
                'compression':
                    self.xml_state.build_type.get_squashfscompression()
            }
        ) as squashed_image:
            squashed_image.create_on_file(squashed_image_file)
        Command.run(
            ['mv', squashed_image_file, self.media_dir.name]
        )

        log.info(
            'Setting up install image bootloader configuration'
        )
        with self._create_bootloader_instance() as bootloader_config:
            bootloader_config.setup_install_boot_images(
                mbrid=self.mbrid,
                lookup_path=self.boot_image_task.boot_root_directory
            )
            IsoToolsBase.setup_media_loader_directory(
                self.boot_image_task.boot_root_directory,
                self.media_dir.name,
                bootloader_config.get_boot_theme()
            )
            bootloader_config.write_meta_data()
            bootloader_config.setup_install_image_config(
                mbrid=self.mbrid
            )
            bootloader_config.write()

            # create initrd for install image
            log.info('Creating install image boot image')
            self._create_iso_install_kernel_and_initrd()

            # the system image initrd is stored to allow kexec
            self._copy_system_image_initrd_to_iso_image()

            if self.firmware.efi_mode():
                efi_loader = Temporary(
                    prefix='efi-loader.', path=self.target_dir
                ).new_file()
                bootloader_config._create_embedded_fat_efi_image(
                    efi_loader.name
                )
                self.custom_iso_args['meta_data']['efi_loader'] = \
                    efi_loader.name

        # create iso filesystem from media_dir
        log.info('Creating ISO filesystem')
        with FileSystemIsoFs(
            device_provider=DeviceProvider(), root_dir=self.media_dir.name,
            custom_args=self.custom_iso_args
        ) as iso_image:
            iso_image.create_on_file(self.isoname)
        self.boot_image_task.cleanup()

    def create_install_pxe_archive(self) -> None:
        """
        Create an oem install tar archive suitable for installing a
        disk image via the network using the PXE boot protocol.
        The archive contains:

        * The raw system image xz compressed
        * The raw system image checksum metadata file
        * The append file template for the boot server
        * The system image initrd for kexec
        * The install initrd
        * The kernel

        Image types which triggers this builder are:

        * installpxe="true|false"
        """
        self.pxe_dir = Temporary(
            prefix='kiwi_pxe_install_media.', path=self.target_dir
        ).new_dir()
        # the system image is transfered as xz compressed variant
        log.info('xz compressing disk image')
        pxe_image_filename = ''.join(
            [
                self.pxe_dir.name, '/',
                self.pxename, '.xz'
            ]
        )
        compress = Compress(
            source_filename=self.diskname,
            keep_source_on_compress=True
        )
        xz_archive = compress.xz(self.xz_options)
        Command.run(
            ['mv', xz_archive, pxe_image_filename]
        )

        # the system image transfer is checked against a checksum
        log.info('Creating disk image checksum')
        pxe_sha256_filename = ''.join(
            [
                self.pxe_dir.name, '/',
                self.pxename, '.sha256'
            ]
        )
        checksum = Checksum(self.diskname)
        checksum.sha256(pxe_sha256_filename)

        # the install image name is stored in a config file
        if self.initrd_system == 'kiwi':
            self._write_install_image_info_to_boot_image()

        # the kexec required system image initrd is stored for dracut kiwi-dump
        if self.initrd_system == 'dracut':
            boot_names = self.boot_image_task.get_boot_names()
            system_image_initrd = os.sep.join(
                [self.root_dir, 'boot', boot_names.initrd_name]
            )
            target_initrd_name = '{0}/{1}.initrd'.format(
                self.pxe_dir.name, self.pxename
            )
            shutil.copy(
                system_image_initrd, target_initrd_name
            )
            os.chmod(target_initrd_name, 420)

        # create pxe config append information
        # this information helps to configure the boot server correctly
        append_filename = ''.join(
            [
                self.pxe_dir.name, '/', self.pxename, '.append'
            ]
        )
        if self.initrd_system == 'kiwi':
            cmdline = 'pxe=1'
        else:
            cmdline = ' '.join(
                [
                    'rd.kiwi.install.pxe',
                    'rd.kiwi.install.image=http://example.com/image.xz'
                ]
            )
        custom_cmdline = self.xml_state.build_type.get_kernelcmdline()
        if custom_cmdline:
            cmdline += ' ' + custom_cmdline
        with open(append_filename, 'w') as append:
            append.write(f'{cmdline}\n')

        # create initrd for pxe install
        log.info('Creating pxe install boot image')
        self._create_pxe_install_kernel_and_initrd()

        # create pxe image bound boot config file, contents can be
        # changed but presence is required.
        log.info('Creating pxe install boot options file')
        configname = '{0}.config.bootoptions'.format(self.pxename)
        shutil.copy(
            os.sep.join([self.root_dir, 'config.bootoptions']),
            os.sep.join([self.pxe_dir.name, configname])
        )

        # create pxe install tarball
        log.info('Creating pxe install archive')
        archive = ArchiveTar(self.pxetarball)

        archive.create(self.pxe_dir.name)
        self.boot_image_task.cleanup()

    def _create_bootloader_instance(self):
        return create_boot_loader_config(
            name=self.bootloader, xml_state=self.xml_state,
            root_dir=self.root_dir,
            boot_dir=self.media_dir.name,
            custom_args=Defaults.get_grub_custom_arguments(
                self.root_dir
            ) if self.bootloader.startswith('grub') else {}
        )

    def _create_pxe_install_kernel_and_initrd(self) -> None:
        kernelname = 'pxeboot.{0}.kernel'.format(self.pxename)
        initrdname = 'pxeboot.{0}.initrd'.format(self.pxename)
        kernel = Kernel(self.boot_image_task.boot_root_directory)
        if kernel.get_kernel():
            kernel.copy_kernel(self.pxe_dir.name, kernelname)
            os.symlink(
                kernelname, ''.join(
                    [
                        self.pxe_dir.name, '/',
                        self.pxename, '.kernel'
                    ]
                )
            )
        else:
            raise KiwiInstallBootImageError(
                'No kernel in boot image tree %s found' %
                self.boot_image_task.boot_root_directory
            )
        if self.xml_state.is_xen_server():
            if kernel.get_xen_hypervisor():
                kernel.copy_xen_hypervisor(
                    self.pxe_dir.name,
                    '/pxeboot.{0}.xen.gz'.format(self.pxename)
                )
            else:
                raise KiwiInstallBootImageError(
                    'No hypervisor in boot image tree %s found' %
                    self.boot_image_task.boot_root_directory
                )
        if self.initrd_system == 'dracut':
            self.boot_image_task.include_module('kiwi-dump')
            self.boot_image_task.include_module('kiwi-dump-reboot')
            self.boot_image_task.omit_module('kiwi-repart')
            if self.root_filesystem_is_multipath is False:
                self.boot_image_task.omit_module('multipath')
            for mod in self.xml_state.get_installmedia_initrd_modules('add'):
                self.boot_image_task.include_module(mod)
            for mod in self.xml_state.get_installmedia_initrd_modules('omit'):
                self.boot_image_task.omit_module(mod)
            self.boot_image_task.set_static_modules(
                self.xml_state.get_installmedia_initrd_modules('set')
            )
            for drv in self.xml_state.get_installmedia_initrd_drivers('add'):
                self.boot_image_task.include_driver(drv)
            for drv in self.xml_state.get_installmedia_initrd_drivers('omit'):
                self.boot_image_task.omit_driver(drv)

        self.boot_image_task.create_initrd(
            self.mbrid, 'initrd_kiwi_install',
            install_initrd=True
        )
        Command.run(
            [
                'mv', self.boot_image_task.initrd_filename,
                self.pxe_dir.name + '/{0}'.format(initrdname)
            ]
        )
        os.chmod(self.pxe_dir.name + '/{0}'.format(initrdname), 420)

    def _create_iso_install_kernel_and_initrd(self) -> None:
        boot_path = self.media_dir.name + '/boot/' + self.arch + '/loader'
        Path.create(boot_path)
        kernel = Kernel(self.boot_image_task.boot_root_directory)
        if kernel.get_kernel():
            kernel.copy_kernel(boot_path, '/linux')
        else:
            raise KiwiInstallBootImageError(
                'No kernel in boot image tree %s found' %
                self.boot_image_task.boot_root_directory
            )
        if self.xml_state.is_xen_server():
            if kernel.get_xen_hypervisor():
                kernel.copy_xen_hypervisor(boot_path, '/xen.gz')
            else:
                raise KiwiInstallBootImageError(
                    'No hypervisor in boot image tree %s found' %
                    self.boot_image_task.boot_root_directory
                )
        if self.initrd_system == 'dracut':
            self.boot_image_task.include_module('kiwi-dump')
            self.boot_image_task.include_module('kiwi-dump-reboot')
            self.boot_image_task.omit_module('kiwi-repart')
            if self.root_filesystem_is_multipath is False:
                self.boot_image_task.omit_module('multipath')
            for mod in self.xml_state.get_installmedia_initrd_modules('add'):
                self.boot_image_task.include_module(mod)
            for mod in self.xml_state.get_installmedia_initrd_modules('omit'):
                self.boot_image_task.omit_module(mod)
            self.boot_image_task.set_static_modules(
                self.xml_state.get_installmedia_initrd_modules('set')
            )

            for drv in self.xml_state.get_installmedia_initrd_drivers('add'):
                self.boot_image_task.include_driver(drv)
            for drv in self.xml_state.get_installmedia_initrd_drivers('omit'):
                self.boot_image_task.omit_driver(drv)

            self._add_system_image_boot_options_to_boot_image()
            self._add_system_identifier_to_boot_image()
        self.boot_image_task.create_initrd(
            self.mbrid, 'initrd_kiwi_install',
            install_initrd=True
        )
        Command.run(
            [
                'mv', self.boot_image_task.initrd_filename,
                boot_path + '/initrd'
            ]
        )

    def _add_system_identifier_to_boot_image(self) -> None:
        filename = ''.join(
            [self.boot_image_task.boot_root_directory, '/system_identifier']
        )
        blockid = BlockID(self.diskname)
        with open(filename, 'w') as system_identifier:
            system_identifier.write(blockid.get_ptuuid())
        self.boot_image_task.include_file(os.sep + os.path.basename(filename))

    def _add_system_image_boot_options_to_boot_image(self) -> None:
        filename = ''.join(
            [self.boot_image_task.boot_root_directory, '/config.bootoptions']
        )
        self.boot_image_task.include_file(os.sep + os.path.basename(filename))

    def _copy_system_image_initrd_to_iso_image(self) -> None:
        boot_names = self.boot_image_task.get_boot_names()
        system_image_initrd = os.sep.join(
            [self.root_dir, 'boot', boot_names.initrd_name]
        )
        shutil.copy(
            system_image_initrd, self.media_dir.name + '/initrd.system_image'
        )

    def _write_install_image_info_to_iso_image(self) -> None:
        iso_trigger = self.media_dir.name + '/config.isoclient'
        with open(iso_trigger, 'w') as iso_system:
            iso_system.write(f'IMAGE="{self.squashed_diskname}\"\n')

    def _write_install_image_info_to_boot_image(self) -> None:
        initrd_trigger = \
            self.boot_image_task.boot_root_directory + '/config.vmxsystem'
        with open(initrd_trigger, 'w') as vmx_system:
            vmx_system.write(f'IMAGE="{self.squashed_diskname}\"\n')



# Copyright (c) 2023 Marcus Schäfer.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import glob
from configparser import ConfigParser
from typing import List, Dict

# project
import kiwi.defaults as defaults
from kiwi.utils.temporary import (
    Temporary, TmpT
)
from kiwi.defaults import Defaults
from kiwi.command import Command
from kiwi.repository.base import RepositoryBase
from kiwi.path import Path
from kiwi.utils.rpm_database import RpmDataBase
from kiwi.utils.toenv import ToEnv


class RepositoryDnf5(RepositoryBase):
    """
    **Implements repository handling for next generation dnf5 package manager**

    :param str shared_dnf_dir: shared directory between image root
        and build system root
    :param str runtime_dnf_config_file: dnf runtime config file name
    :param dict command_env: customized os.environ for dnf
    :param str runtime_dnf_config: instance of :class:`ConfigParser`
    """

    def post_init(self, custom_args: List = []) -> None:
        """
        Post initialization method

        Store custom dnf arguments and create runtime configuration
        and environment

        :param list custom_args: dnf arguments
        """
        self.runtime_dnf_config_file = TmpT(name='')
        self.custom_args = custom_args
        self.exclude_docs = False
        self.locale = None
        self.target_arch = None

        # extract custom arguments not used in dnf call
        if 'exclude_docs' in self.custom_args:
            self.custom_args.remove('exclude_docs')
            self.exclude_docs = True

        if 'check_signatures' in self.custom_args:
            self.custom_args.remove('check_signatures')
            self.gpg_check = '1'
        else:
            self.gpg_check = '0'

        for argument in self.custom_args:
            if '_install_langs' in argument:
                self.locale = argument
            elif '_target_arch' in argument:
                self.target_arch = argument
        if self.locale:
            self.custom_args.remove(self.locale)
        if self.target_arch:
            self.custom_args.remove(self.target_arch)
            self.target_arch = self.target_arch.split('%')[1]

        self.repo_names: List = []

        # dnf support is based on creating repo files which contains
        # path names to the repo and its cache. In order to allow a
        # persistent use of the files in and outside of a chroot call
        # an active bind mount from RootBind::mount_shared_directory
        # is expected and required
        manager_base = self.shared_location + '/dnf'

        self.shared_dnf_dir = {
            'reposd-dir': manager_base + '/repos',
            'cache-dir': manager_base + '/cache',
            'pluginconf-dir': manager_base + '/pluginconf',
            'vars-dir': manager_base + '/vars'
        }

        self.runtime_dnf_config_file = Temporary(
            path=self.root_dir, prefix='kiwi_dnf5.config'
        ).unmanaged_file()

        self.dnf_args = [
            '--config', self.runtime_dnf_config_file.name, '-y', '--disable-plugin=priorities,versionlock'
        ] + self.custom_args

        self.command_env = self._create_dnf_runtime_environment()

        # config file parameters for dnf tool
        self._create_runtime_config_parser()
        self._write_runtime_config()

    def setup_package_database_configuration(self) -> None:
        """
        Setup rpm macros for bootstrapping and image building

        1. Create the rpm image macro which persists during the build
        2. Create the rpm bootstrap macro to make sure for bootstrapping
           the rpm database location matches the host rpm database setup.
           This macro only persists during the bootstrap phase. If the
           image was already bootstrapped a compat link is created instead.
        """
        rpmdb = RpmDataBase(
            self.root_dir, Defaults.get_custom_rpm_image_macro_name()
        )
        if self.locale:
            rpmdb.set_macro_from_string(self.locale)
        rpmdb.write_config()

        rpmdb = RpmDataBase(self.root_dir)
        if rpmdb.has_rpm():
            rpmdb.link_database_to_host_path()
        else:
            rpmdb.set_database_to_host_path()
        # SUSE compat code:
        #
        # Manually adding the compat link /var/lib/rpm that points to the
        # rpmdb location as it is configured in the host rpm setup. DNF makes
        # use of the host setup to bootstrap or to read the signing keys
        # from the rpm database setup provisioned by rpm itself (macro level).
        # It assumes the host setup is the default for the system being built.
        #
        # For SUSE this is causing a mismatch as the host setup is not the RPM
        # default (/var/lib/rpm) and SUSE distros proactively relocate the rpm
        # database from the default location to a custom one when the
        # rpm package and related macros are installed for the first time.
        rpmdb.init_database()
        image_rpm_compat_link = '/var/lib/rpm'
        host_rpm_dbpath = rpmdb.rpmdb_host.expand_query('%_dbpath')
        if host_rpm_dbpath != image_rpm_compat_link:
            Path.create(
                self.root_dir + os.path.dirname(image_rpm_compat_link)
            )
            Command.run(
                [
                    'ln', '-s', '--no-target-directory',
                    ''.join(['../..', host_rpm_dbpath]),
                    self.root_dir + image_rpm_compat_link
                ], raise_on_error=False
            )

    def use_default_location(self) -> None:
        """
        Setup dnf repository operations to store all data
        in the default places
        """
        self.shared_dnf_dir['reposd-dir'] = \
            self.root_dir + '/etc/yum.repos.d'
        self.shared_dnf_dir['cache-dir'] = \
            self.root_dir + '/var/cache/libdnf5'
        self.shared_dnf_dir['pluginconf-dir'] = \
            self.root_dir + '/etc/dnf/libdnf5-plugins'
        self.shared_dnf_dir['vars-dir'] = \
            self.root_dir + '/etc/dnf/vars'
        self._create_runtime_config_parser()
        self._write_runtime_config()

    def runtime_config(self) -> Dict:
        """
        dnf runtime configuration and environment

        :return: dnf_args:list, command_env:dict

        :rtype: dict
        """
        return {
            'dnf_args': self.dnf_args,
            'command_env': self.command_env
        }

    def add_repo(
        self, name: str, uri: str, repo_type: str = 'rpm-md',
        prio: int = None, dist: str = None, components: str = None,
        user: str = None, secret: str = None, credentials_file: str = None,
        repo_gpgcheck: bool = False, pkg_gpgcheck: bool = False,
        sourcetype: str = None, customization_script: str = None,
        architectures: str = None
    ) -> None:
        """
        Add dnf repository

        :param str name: repository base file name
        :param str uri: repository URI
        :param str repo_type: repostory type name
        :param int prio: dnf repostory priority
        :param str dist: unused
        :param str components: unused
        :param str user: unused
        :param str secret: unused
        :param str credentials_file: unused
        :param bool repo_gpgcheck: enable repository signature validation
        :param bool pkg_gpgcheck: enable package signature validation
        :param str sourcetype:
            source type, one of 'baseurl', 'metalink' or 'mirrorlist'
        :param str customization_script:
            custom script called after the repo file was created
        :param str architectures: unused
        """
        repo_file = self.shared_dnf_dir['reposd-dir'] + '/' + name + '.repo'
        self.repo_names.append(name + '.repo')
        if os.path.exists(uri):
            # dnf requires local paths to take the file: type
            uri = 'file://' + uri
        repo_config = ConfigParser(interpolation=None)
        repo_config.add_section(name)
        repo_config.set(
            name, 'name', name
        )
        repo_config.set(
            name, sourcetype if sourcetype else 'baseurl', uri
        )
        if prio:
            repo_config.set(
                name, 'priority', format(prio)
            )
        repo_config.set(
            name, 'repo_gpgcheck', '1' if repo_gpgcheck else '0'
        )
        repo_config.set(
            name, 'gpgcheck', '1' if pkg_gpgcheck else '0'
        )
        if Defaults.is_buildservice_worker():
            # when building in the build service, modular metadata is
            # inaccessible. In order to use modular content in the build
            # service, we need to disable modular filtering, which is
            # done with module_hotfixes option
            repo_config.set(
                name, 'module_hotfixes', '1'
            )
        if not os.path.isdir(self.shared_dnf_dir['reposd-dir']):
            Path.create(self.shared_dnf_dir['reposd-dir'])
        with open(repo_file, 'w') as repo:
            repo_config.write(repo)
        if customization_script:
            self.run_repo_customize(customization_script, repo_file)

    def import_trusted_keys(self, signing_keys: List) -> None:
        """
        Imports trusted keys into the image

        :param list signing_keys: list of the key files to import
        """
        rpmdb = RpmDataBase(self.root_dir)
        for key in signing_keys:
            rpmdb.import_signing_key_to_image(key)

    def delete_repo(self, name: str) -> None:
        """
        Delete dnf repository

        :param str name: repository base file name
        """
        Path.wipe(
            self.shared_dnf_dir['reposd-dir'] + '/' + name + '.repo'
        )

    def delete_all_repos(self) -> None:
        """
        Delete all dnf repositories
        """
        Path.wipe(self.shared_dnf_dir['reposd-dir'])
        Path.create(self.shared_dnf_dir['reposd-dir'])

    def delete_repo_cache(self, name: str) -> None:
        """
        Delete dnf repository cache

        The cache data for each repository is stored in a directory
        and additional files all starting with the repository name.
        The method glob deletes all files and directories matching
        the repository name followed by any characters to cleanup
        the cache information

        :param str name: repository name
        """
        dnf_cache_glob_pattern = ''.join(
            [self.shared_dnf_dir['cache-dir'], os.sep, name, '*']
        )
        for dnf_cache_file in glob.iglob(dnf_cache_glob_pattern):
            Path.wipe(dnf_cache_file)

    def cleanup_unused_repos(self) -> None:
        """
        Delete unused dnf repositories

        Repository configurations which are not used for this build
        must be removed otherwise they are taken into account for
        the package installations
        """
        repos_dir = self.shared_dnf_dir['reposd-dir']
        repo_files = list(os.walk(repos_dir))[0][2]
        for repo_file in repo_files:
            if repo_file not in self.repo_names:
                Path.wipe(repos_dir + '/' + repo_file)

    def _create_dnf_runtime_environment(self) -> Dict:
        for dnf_dir in list(self.shared_dnf_dir.values()):
            Path.create(dnf_dir)
        ToEnv(self.root_dir, defaults.PACKAGE_MANAGER_ENV_VARS)
        return dict(
            os.environ, LANG='C'
        )

    def _create_runtime_config_parser(self) -> None:
        self.runtime_dnf_config = ConfigParser(interpolation=None)
        self.runtime_dnf_config["main"] = {
            "system_cachedir": self.shared_dnf_dir['cache-dir'],
            "reposdir": self.shared_dnf_dir['reposd-dir'],
            "varsdir": self.shared_dnf_dir['vars-dir'],
            "pluginconfpath": self.shared_dnf_dir['pluginconf-dir'],
            "keepcache": "1",
            "debuglevel": "2",
            "best": "1",
            "obsoletes": "1",
            "plugins": "0",
            "gpgcheck": self.gpg_check,
        }
        if self.exclude_docs:
            self.runtime_dnf_config["main"]["tsflags"] = "nodocs"
        if self.target_arch:
            self.runtime_dnf_config["main"].update({
                "arch": self.target_arch,
                "ignorearch": "1",
            })

    def _write_runtime_config(self) -> None:
        with open(self.runtime_dnf_config_file.name, 'w') as config:
            self.runtime_dnf_config.write(config)

    def cleanup(self) -> None:
        """
        Delete intermediate dnf config file
        """
        if os.path.isfile(self.runtime_dnf_config_file.name):
            os.unlink(self.runtime_dnf_config_file.name)



# Copyright (c) 2025 SUSE Software Solutions Germany GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
from typing import (
    List, Dict, Optional, Any
)

# project
from kiwi.mount_manager import MountManager


class SnapshotManagerBase:
    """
    **Implements base class for snapshots management inferface**

    :param str device: storage device node name
    :param str root_dir: root directory path
    :param str mountpoint: mountpoint of the filesystem to snapshot
    :param str root_volume_name: the name of the root volume in case
        snapshots are hosted in a subvolume.
    :param dict custom_args: dictionary of custom snapshot manager arguments
    """
    def __init__(
        self, device: str, root_dir: str, mountpoint: str,
        root_volume_name: str, custom_args: Optional[Dict[str, Any]] = None
    ) -> None:
        self.device = device
        self.root_dir = root_dir
        self.mountpoint = mountpoint
        self.root_volume_name = root_volume_name
        self.post_init(custom_args)

    def post_init(self, custom_args):
        """
        Post initialization method

        Implementation in specialized SnapshotManager class if required
        """
        pass

    def create_first_snapshot(self) -> List[MountManager]:
        raise NotImplementedError

    def setup_first_snapshot(self):
        raise NotImplementedError

    def get_default_snapshot_name(self) -> str:
        raise NotImplementedError



# Copyright (c) 2025 SUSE LLC.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os

# project
from kiwi.container.setup.base import ContainerSetupBase
from kiwi.exceptions import KiwiContainerSetupError


class ContainerSetupWsl(ContainerSetupBase):
    """
    WSL new tar setup
    """
    def setup(self):
        """
        Setup container metadata
        """
        # It's expected that the container metadata for the
        # new WSL format is provided by a package or via overlay
        # data. Thus this setup routine only checks for the
        # presence of the data and fails when missing
        distribution_conf = os.path.normpath(
            os.sep.join([self.root_dir, 'etc', 'wsl-distribution.conf'])
        )
        wsl_conf = os.path.normpath(
            os.sep.join([self.root_dir, 'etc', 'wsl.conf'])
        )
        if not os.path.exists(distribution_conf):
            raise KiwiContainerSetupError(
                f'Mandatory WSL {distribution_conf} not found in root tree'
            )
        if not os.path.exists(wsl_conf):
            raise KiwiContainerSetupError(
                f'Mandatory WSL {wsl_conf} not found in root tree'
            )



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import logging
import pathlib
from typing import (
    Dict, List
)

# project
from kiwi.system.root_import.base import RootImportBase
from kiwi.mount_manager import MountManager
from kiwi.path import Path
from kiwi.defaults import Defaults
from kiwi.utils.compress import Compress
from kiwi.oci_tools import OCI

log = logging.getLogger('kiwi')


class RootImportOCI(RootImportBase):
    """
    Implements the base class for importing a root system from
    a oci image tarball file.
    """
    def post_init(self, custom_args: Dict[str, str]):
        self.archive_transport = custom_args['archive_transport']
        self.compressed_image_files: List[Compress] = []

    def sync_data(self):
        """
        Synchronize data from the given base image to the target root
        directory.
        """
        for image_url in self._get_image_urls():
            with OCI.new() as oci:
                oci.import_container_image(image_url)
                oci.unpack()
                oci.import_rootfs(self.root_dir)

        # A copy of the uncompressed image(all derived imports) and
        # its checksum are kept inside the root_dir in order to ensure
        # the later steps i.e. system create are atomic and don't need
        # any third party archive.
        image_copy = Defaults.get_imported_root_image(self.root_dir)

        Path.create(os.path.dirname(image_copy))
        oci.export_container_image(
            image_copy, 'oci-archive',
            Defaults.get_container_base_image_tag()
        )
        self._make_checksum(image_copy)

    def overlay_data(self) -> None:
        """
        Synchronize data from the given base image to the target root
        directory as an overlayfs mounted target.
        """
        root_dir_ro = f'{self.root_dir}_ro'

        for image_url in self._get_image_urls():
            with OCI.new() as oci:
                oci.import_container_image(image_url)
                oci.unpack()
                oci.import_rootfs(self.root_dir)

        log.debug("renaming %s -> %s", self.root_dir, root_dir_ro)
        pathlib.Path(self.root_dir).replace(root_dir_ro)
        Path.create(self.root_dir)

        self.overlay = MountManager(device='', mountpoint=self.root_dir)
        self.overlay.overlay_mount(root_dir_ro)

    def _get_image_urls(self) -> List[str]:
        if not self.raw_urls:
            image_urls: List[str] = []
            for image_file in self.image_files:
                compressor = Compress(image_file)
                # Increase livetime of the the compressor instances
                # to the livetime of RootImportOCI. They create
                # temporary files which are referenced later and
                # need to live longer than this loop block
                self.compressed_image_files.append(compressor)
                if compressor.get_format():
                    compressor.uncompress(True)
                    uncompressed_image = compressor.uncompressed_filename
                else:
                    uncompressed_image = image_file
                image_urls.append(
                    '{0}:{1}'.format(
                        self.archive_transport, uncompressed_image
                    )
                )
            return image_urls
        else:
            log.warning('Bypassing base image URI to OCI tools')
            return self.raw_urls



# Copyright (c) 2025 SUSE Software Solutions Germany GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import shutil
import datetime
from typing import List
from xml.etree import ElementTree
from xml.dom import minidom

# project
from kiwi.command import Command
from kiwi.utils.command_capabilities import CommandCapabilities
from kiwi.path import Path
from kiwi.defaults import Defaults
from kiwi.snapshot_manager.base import SnapshotManagerBase
from kiwi.mount_manager import MountManager
from kiwi.utils.sysconfig import SysConfig
from kiwi.chroot_manager import (
    ChrootManager, ChrootMount
)
from kiwi.exceptions import (
    KiwiSnapshotManagerSetupError
)


class SnapshotManagerSnapper(SnapshotManagerBase):
    """
    **Implements the snapshots management for Snapper**

    :param str device: storage device node name
    :param str root_dir: root directory path
    :param str mountpoint: mountpoint of the filesystem to snapshot
    :param str root_volume_name: the name of the root volume in case
        snapshots are hosted in a subvolume.
    """
    def post_init(self, custom_args):
        """
        Post initialization method

        Store custom snapper manager initialization arguments

        :param dict custom_args: custom snapper manager arguments
        """
        self.default_snapshot_name = \
            f'/{self.root_volume_name}/.snapshots/1/snapshot'
        if custom_args:
            self.custom_args = custom_args
        else:
            self.custom_args = {}
        if 'quota_groups' not in self.custom_args:
            self.custom_args['quota_groups'] = False

    def create_first_snapshot(self) -> List[MountManager]:
        """
        Creates the first snapshot on an empty filesystem

        In fact it is essentially creating an empty subvolume
        following the snapper's structure and convetion to store
        snapshots

        :return: list of additional mountpoints added on top of the given
            root mountpoint

        :rtype: list
        """
        snapshot_path = os.path.normpath(os.sep.join([
            self.mountpoint,
            self.default_snapshot_name
        ]))
        snapshot_path = self.mountpoint + self.default_snapshot_name
        if CommandCapabilities.check_version(
            'snapper', (0, 12, 1), root=self.root_dir
        ):
            with ChrootManager(
                self.root_dir, binds=[ChrootMount(self.mountpoint)]
            ) as chroot:
                chroot.run([
                    '/usr/lib/snapper/installation-helper', '--root-prefix',
                    os.sep.join([self.mountpoint, self.root_volume_name]),
                    '--step', 'filesystem'
                ])
        else:
            # with snapper <= 0.12.0 the first snapshot subvolume
            # structure is crafted manually
            snapshot_volume = self.mountpoint + \
                f'/{self.root_volume_name}/.snapshots'
            Command.run(
                ['btrfs', 'subvolume', 'create', snapshot_volume]
            )
            os.chmod(snapshot_volume, 0o700)
            Path.create(snapshot_volume + '/1')
            Command.run(
                ['btrfs', 'subvolume', 'create', snapshot_path]
            )
            Command.run(
                ['btrfs', 'subvolume', 'set-default', snapshot_path]
            )

        # Mount /{some-name}/.snapshots as /.snapshots inside the root
        snapshots_mount = MountManager(
            device=self.device,
            attributes={
                'subvol_path': f'{self.root_volume_name}/.snapshots',
                'subvol_name': f'{self.root_volume_name}/.snapshots'
            },
            mountpoint=snapshot_path + '/.snapshots'
        )
        return [snapshots_mount]

    def setup_first_snapshot(self):
        """
        Arranges the snapshot configuration once the OS tree is synched.
        It uses the snapper and the data from the snapshot itself.
        """
        snapshot_path = os.path.normpath(os.sep.join([
            self.mountpoint,
            self.default_snapshot_name
        ]))
        if CommandCapabilities.check_version(
            'snapper', (0, 12, 1), root=snapshot_path
        ):
            snapshots_prefix = os.sep.join([snapshot_path, '.snapshots'])
            with ChrootManager(
                self.root_dir, binds=[
                    ChrootMount(snapshot_path), ChrootMount(snapshots_prefix)
                ]
            ) as chroot:
                chroot.run([
                    '/usr/lib/snapper/installation-helper', '--root-prefix',
                    snapshot_path, '--step', 'config', '--description',
                    'first root filesystem'
                ])
        else:
            # with snapper <= 0.12.0 the snapshot configuration is
            # crafted manually
            date_info = datetime.datetime.now()
            snapshot = ElementTree.Element('snapshot')

            snapshot_type = ElementTree.SubElement(snapshot, 'type')
            snapshot_type.text = 'single'

            snapshot_number = ElementTree.SubElement(snapshot, 'num')
            snapshot_number.text = '1'

            snapshot_description = ElementTree.SubElement(snapshot, 'description')
            snapshot_description.text = 'first root filesystem'

            snapshot_date = ElementTree.SubElement(snapshot, 'date')
            snapshot_date.text = date_info.strftime("%Y-%m-%d %H:%M:%S")

            with open(os.path.join(snapshot_path, '../info.xml'), 'w') as snapshot_info_file:
                snapshot_info_file.write(self._xml_pretty(snapshot))

        if self.custom_args['quota_groups']:
            snapper_default_conf = Defaults.get_snapper_config_template_file(
                snapshot_path
            )
            if snapper_default_conf:
                # snapper requires an extra parent qgroup to operate with quotas
                Command.run(
                    ['btrfs', 'qgroup', 'create', '1/0', self.mountpoint]
                )
                config_file = self._set_snapper_sysconfig_file(snapshot_path)
                if not os.path.exists(config_file):
                    shutil.copyfile(snapper_default_conf, config_file)
                with ChrootManager(snapshot_path) as chroot:
                    chroot.run([
                        'snapper', '--no-dbus', 'set-config', 'QGROUP=1/0'
                    ])
        with ChrootManager(snapshot_path) as chroot:
            chroot.run([
                'snapper', '--no-dbus', 'modify', '--default', '1'
            ])

    def get_default_snapshot_name(self) -> str:
        """
        Gets the name of the default snapshot for snapper

        :return: default snapper snapshot path

        :rtype: string
        """
        return self.default_snapshot_name

    def _xml_pretty(self, toplevel_element):
        xml_data_unformatted = ElementTree.tostring(
            toplevel_element, 'utf-8'
        )
        xml_data_domtree = minidom.parseString(xml_data_unformatted)
        return xml_data_domtree.toprettyxml(indent="    ")

    @staticmethod
    def _set_snapper_sysconfig_file(root_path):
        sysconf_file = SysConfig(
            os.sep.join([root_path, 'etc/sysconfig/snapper'])
        )
        if not sysconf_file.get('SNAPPER_CONFIGS') or \
           len(sysconf_file['SNAPPER_CONFIGS'].strip('\"')) == 0:

            sysconf_file['SNAPPER_CONFIGS'] = '"root"'
            sysconf_file.write()
        elif len(sysconf_file['SNAPPER_CONFIGS'].split()) > 1:
            raise KiwiSnapshotManagerSetupError(
                'Unsupported SNAPPER_CONFIGS value: {0}'.format(
                    sysconf_file['SNAPPER_CONFIGS']
                )
            )
        return os.sep.join([
            root_path, 'etc/snapper/configs',
            sysconf_file['SNAPPER_CONFIGS'].strip('\"')]
        )



# Copyright (c) 2025 SUSE Software Solutions Germany GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import logging
from typing import (
    List, Optional, NamedTuple
)

# project
from kiwi.mount_manager import MountManager
from kiwi.command import (
    Command, CommandT, MutableMapping
)
from kiwi.exceptions import (
    KiwiUmountBusyError
)

log = logging.getLogger('kiwi')


class ChrootMount(NamedTuple):
    target: str
    source: Optional[str] = None


class ChrootManager:
    """
    **Implements methods for setting and unsetting a chroot environment**

    The caller is responsible for cleaning up bind mounts if the ChrootManager
    is used as is, without a context.

    The class also supports to be used as a context manager, where any bind or kernel
    filesystem mount is unmounted once the context manager's with block is left

    * :param string root_dir: path to change the root to
    * :param list binds: current root paths to bind to the chrooted path
    """
    def __init__(self, root_dir: str, binds: List[ChrootMount] = []):
        self.root_dir = root_dir
        self.mounts: List[MountManager] = []
        for bind in binds:
            self.mounts.append(MountManager(
                device=bind.source if bind.source else bind.target,
                mountpoint=os.path.normpath(
                    os.sep.join([root_dir, bind.target])
                )
            ))

    def __enter__(self) -> "ChrootManager":
        try:
            self.mount()
        except Exception as e:
            try:
                self.umount()
            except Exception:
                pass
            raise e
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.umount()

    def mount(self) -> None:
        """
        Mounts binds to the chroot path
        """
        for mnt in self.mounts:
            mnt.bind_mount()

    def umount(self) -> None:
        """
        Unmounts all binds from the chroot path

        If any unmount raises a KiwiUmountBusyError this is trapped
        and kept until the iteration over all bind mounts is over.
        """
        errors = []
        for mnt in reversed(self.mounts):
            try:
                mnt.umount()
            except KiwiUmountBusyError as e:
                errors.append(e)

        if errors:
            raise KiwiUmountBusyError(errors)

    def run(
        self, command: List[str],
        custom_env: Optional[MutableMapping[str, str]] = None,
        raise_on_error: bool = True, stderr_to_stdout: bool = False,
        raise_on_command_not_found: bool = True
    ) -> Optional[CommandT]:
        """
        This is a wrapper for Command.run method but pre-appending the
        chroot call at the command list

        :param list command: command and arguments
        :param dict custom_env: custom os.environ
        :param bool raise_on_error: control error behaviour
        :param bool stderr_to_stdout: redirects stderr to stdout

        :return:
            Contains call results in command type

            .. code:: python

                CommandT(output='string', error='string', returncode=int)

        :rtype: CommandT
        """
        chroot_cmd = ['chroot', self.root_dir]
        chroot_cmd = chroot_cmd + command
        return Command.run(
            chroot_cmd, custom_env, raise_on_error, stderr_to_stdout,
            raise_on_command_not_found
        )



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import logging
import os
import random
import struct

# project
from kiwi.storage.device_provider import DeviceProvider

log = logging.getLogger('kiwi')


class SystemIdentifier:
    """
    **Create a random ID to identify the system**

    The information is used to create the mbrid file as an example

    :param str image_id: hex identifier string
    """
    def __init__(self):
        self.image_id = None
        sde = os.environ.get('SOURCE_DATE_EPOCH')
        if sde:
            log.info(f'Using SystemIdentifier seed SOURCE_DATE_EPOCH:{sde}')
            random.seed(int(sde))

    def get_id(self) -> str:
        """
        Current hex identifier

        :return: hex id

        :rtype: str
        """
        return self.image_id

    def calculate_id(self) -> None:
        """
        Calculate random hex id

        Using 4 tuples of rand in range from 1..0xfe
        """
        self.image_id = '0x%02x%02x%02x%02x' % (
            self._rand(), self._rand(), self._rand(), self._rand()
        )

    def write(self, filename: str) -> None:
        """
        Write current hex identifier to file

        :param str filename: file path name
        """
        with open(filename, 'w') as identifier:
            identifier.write(f'{self.image_id}\n')

    def write_to_disk(self, device_provider: DeviceProvider) -> None:
        """
        Write current hex identifier to MBR at offset 0x1b8 on disk

        :param object device_provider: Instance based on DeviceProvider
        """
        if self.image_id:
            packed_id = struct.pack('<I', int(self.image_id, 16))
            with open(device_provider.get_device(), 'wb') as disk:
                disk.seek(440, 0)
                disk.write(packed_id)

    def _rand(self):
        return random.randrange(1, 0xfe)



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import importlib
from typing import Dict, Optional, Tuple

# project
from kiwi.storage.subformat.base import DiskFormatBase
from kiwi.xml_state import XMLState
from kiwi.exceptions import KiwiDiskFormatSetupError


class DiskFormat:
    """
    **DiskFormat factory**

    :param string name: Format name
    :param object xml_state: Instance of XMLState
    :param string root_dir: root directory path name
    :param string target_dir: target directory path name
    """

    @staticmethod
    def new(
        name: str, xml_state: XMLState, root_dir: str, target_dir: str
    ) -> DiskFormatBase:
        name_map = {
            'oci': 'Oci',
            'qcow2': 'Qcow2',
            'vdi': 'Vdi',
            'vhd': 'Vhd',
            'vhdx': 'Vhdx',
            'vhdfixed': 'VhdFixed',
            'gce': 'Gce',
            'vmdk': 'Vmdk',
            'ova': 'Ova',
            'vagrant_libvirt': 'VagrantLibVirt',
            'vagrant_virtualbox': 'VagrantVirtualBox',
            'base': 'Base'
        }
        module_namespace: Optional[str] = None
        try:
            custom_args, module_namespace = DiskFormat.\
                _custom_args_for_format(name, xml_state)
            diskformat = importlib.import_module(
                f'kiwi.storage.subformat.{module_namespace}'
            )
            module_name = f'DiskFormat{name_map[module_namespace]}'
            return diskformat.__dict__[module_name](
                xml_state, root_dir, target_dir, custom_args
            )
        except Exception as issue:
            raise KiwiDiskFormatSetupError(
                f'No support for {module_namespace} disk format: {issue}'
            )

    @staticmethod
    def _custom_args_for_format(name: str, xml_state: XMLState) -> Tuple[Dict, str]:
        custom_args = xml_state.get_build_type_format_options()
        module_namespace = name
        if name == 'vhd-fixed':
            disk_tag = xml_state.build_type.get_vhdfixedtag()
            if disk_tag:
                custom_args.update(
                    {'--tag': disk_tag}
                )
            module_namespace = 'vhdfixed'
        elif name == 'gce':
            gce_license_tag = xml_state.build_type.get_gcelicense()
            if gce_license_tag:
                custom_args.update(
                    {'--tag': gce_license_tag}
                )
        elif name == 'vmdk' or name == 'ova':
            vmdisk_section = xml_state.get_build_type_vmdisk_section()
            if vmdisk_section:
                disk_mode = vmdisk_section.get_diskmode()
                disk_controller = vmdisk_section.get_controller()
                if disk_mode:
                    custom_args.update(
                        {'subformat={0}'.format(disk_mode): None}
                    )
                if disk_controller:
                    custom_args.update(
                        {'adapter_type={0}'.format(disk_controller): None}
                    )
        elif name == 'vagrant':
            vagrant_config = xml_state.get_build_type_vagrant_config_section()
            if vagrant_config:
                custom_args.update(
                    {'vagrantconfig': vagrant_config}
                )
                module_namespace = '{0}_{1}'.format(
                    name, vagrant_config.get_provider()
                )
        elif name == 'raw':
            module_namespace = 'base'
        elif name.startswith('oci:') or name.startswith('docker:'):
            container_format, base_format = name.split(':', 1)
            module_namespace = 'oci'
            custom_args.update(
                {
                    'base_format': base_format,
                    'container_format': container_format
                }
            )
        return custom_args, module_namespace



# Copyright (c) 2015 SUSE Linux GmbH.  All rights reserved.
#
# This file is part of kiwi.
#
# kiwi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# kiwi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with kiwi.  If not, see <http://www.gnu.org/licenses/>
#
import os
import uuid
import random
import logging
import copy
from functools import reduce
from typing import (
    Dict, List, Optional
)

# project
import kiwi.defaults as defaults

from kiwi.defaults import Defaults
from kiwi.utils.sync import DataSync
from kiwi.mount_manager import MountManager
from kiwi.command import Command
from kiwi.storage.device_provider import DeviceProvider
from kiwi.utils.veritysetup import VeritySetup

from kiwi.exceptions import (
    KiwiFileSystemSyncError
)

log = logging.getLogger('kiwi')


class FileSystemBase:
    """
    **Implements base class for filesystem interface**

    :param object device_provider:
        Instance of a class based on DeviceProvider
        required for filesystems which needs a block device for
        creation. In most cases the DeviceProvider is a LoopDevice
    :param string root_dir: root directory path name
    :param dict custom_args: custom filesystem arguments
    """
    def __init__(
        self, device_provider: DeviceProvider,
        root_dir: str = '', custom_args: Dict = {}
    ):
        # filesystems created with a block device stores the mountpoint
        # here. The file name of the file containing the filesystem is
        # stored in the device_provider if the filesystem is represented
        # as a file there
        self.filesystem_mount: Optional[MountManager] = None

        # the underlaying device provider. This is only required if the
        # filesystem required a block device to become created
        self.device_provider = device_provider

        self.root_dir = root_dir

        # filesystems created without a block device stores the result
        # filesystem file name here
        self.filename = ''

        self.custom_args: Dict = {}
        self.post_init(custom_args)
        self.veritysetup: Optional[VeritySetup] = None

    def __enter__(self):
        return self

    def post_init(self, custom_args: Dict):
        """
        Post initialization method

        Store dictionary of custom arguments if not empty. This
        overrides the default custom argument hash

        :param dict custom_args: custom arguments

            .. code:: python

                {
                    'create_options': ['option'],
                    'mount_options': ['option'],
                    'meta_data': {
                        'key': 'value'
                    }
                }
        """
        if custom_args:
            self.custom_args = copy.deepcopy(custom_args)

        if not self.custom_args.get('create_options'):
            self.custom_args['create_options'] = []

        if not self.custom_args.get('meta_data'):
            self.custom_args['meta_data'] = {}

        if not self.custom_args.get('mount_options'):
            self.custom_args['mount_options'] = []

        if not self.custom_args.get('fs_attributes'):
            self.custom_args['fs_attributes'] = []

    def set_uuid(self):
        """
        Create new random filesystem UUID

        Implement in specialized filesystem class for filesystems which
        supports the concept of an UUID and allows to change it
        """
        log.warning(
            'Instance {0} has no support for setting a new UUID label'.format(
                type(self).__name__
            )
        )

    def create_on_device(
        self, label: str = None, size: int = 0, unit: str = defaults.UNIT.kb,
        uuid: str = None
    ):
        """
        Create filesystem on block device

        Implement in specialized filesystem class for filesystems which
        requires a block device for creation, e.g ext4.

        :param str label: label name
        :param int size:
            size value, can also be counted from the end via -X
            The value is interpreted in units of: unit
        :param str unit:
            unit name. Default unit is set to: defaults.UNIT.kb
        :param str uuid: UUID name
        """
        raise NotImplementedError

    def create_on_file(
        self, filename: str, label: str = None, exclude: List[str] = None
    ):
        """
        Create filesystem from root data tree

        Implement in specialized filesystem class for filesystems which
        requires a data tree for creation, e.g squashfs.

        :param string filename: result file path name
        :param string label: label name
        :param list exclude: list of exclude dirs/files
        """
        raise NotImplementedError

    def get_mountpoint(self) -> Optional[str]:
        """
        Provides mount point directory

        Effective use of the directory is guaranteed only after sync_data

        :return: directory path name

        :rtype: string
        """
        if self.filesystem_mount:
            return self.filesystem_mount.mountpoint
        return None

    def sync_data(self, exclude: List[str] = []) -> MountManager:
        """
        Copy data tree into filesystem

        :param list exclude: list of exclude dirs/files
        :return: The mount created for syncing data. It should be used to
            un-mount the filesystem again.
        """
        if not self.root_dir:
            raise KiwiFileSystemSyncError(
                'no root directory specified'
            )
        if not os.path.exists(self.root_dir):
            raise KiwiFileSystemSyncError(
                f'given root directory {self.root_dir} does not exist'
            )
        self.filesystem_mount = MountManager(
            device=self.device_provider.get_device()
        )
        self.filesystem_mount.mount(
            self.custom_args['mount_options']
        )
        self._apply_attributes()
        data = DataSync(
            self.root_dir, self.filesystem_mount.mountpoint
        )
        data.sync_data(
            exclude=exclude, options=Defaults.get_sync_options()
        )
        return self.filesystem_mount

    def create_verity_layer(
        self, blocks: Optional[int] = None, filename: str = None
    ):
        """
        Create veritysetup on device

        :param int block:
            Number of blocks to use for veritysetup.
            If not specified the entire root device is used
        :param str filename:
            Target filename to use for VeritySetup.
            If not specified the filename or block special
            provided at object construction time is used
        """
        on_file_name = filename or self.filename
        self.veritysetup = VeritySetup(
            on_file_name or self.device_provider.get_device(),
            blocks
        )
        log.info(
            '--> Creating dm verity hash ({0} blocks)...'.format(
                blocks or 'all'
            )
        )
        log.debug(
            '--> dm verity metadata: {0}'.format(self.veritysetup.format())
        )

    def create_verification_metadata(self, device_node: str = '') -> None:
        """
        Write verification block at the end of the device

        :param str device_node:
            Target device node, if not specified the root device
            from this instance is used
        """
        if self.veritysetup:
            log.info('--> Creating verification metadata...')
            self.veritysetup.create_verity_verification_metadata()
            log.info('--> Signing verification metadata...')
            self.veritysetup.sign_verification_metadata()
            self.veritysetup.write_verification_metadata(
                device_node or self.device_provider.get_device()
            )

    def mount(self) -> None:
        """
        Mount the filesystem
        """
        if self.filesystem_mount:
            self.filesystem_mount.mount(
                self.custom_args['mount_options']
            )

    def umount(self) -> None:
        """
        Umounts the filesystem in case it is mounted, does nothing otherwise
        """
        if self.filesystem_mount:
            log.info('umount %s instance', type(self).__name__)
            self.filesystem_mount.umount()

    def umount_volumes(self) -> None:
        """
        Consistency layer with regards to VolumeManager classes

        Invokes umount
        """
        self.umount()

    def mount_volumes(self) -> None:
        """
        Consistency layer with regards to VolumeManager classes

        Invokes mount
        """
        self.mount()

    def get_volumes(self) -> Dict:
        """
        Consistency layer with regards to VolumeManager classes

        Raises
        """
        raise NotImplementedError

    def get_root_volume_name(self) -> None:
        """
        Consistency layer with regards to VolumeManager classes

        Raises
        """
        raise NotImplementedError

    def get_fstab(
        self, persistency_type: str = 'by-label', filesystem_name: str = ''
    ) -> List[str]:
        """
        Consistency layer with regards to VolumeManager classes

        Raises
        """
        raise NotImplementedError

    def set_property_readonly_root(self) -> None:
        """
        Consistency layer with regards to VolumeManager classes

        Raises
        """
        raise NotImplementedError

    def _generate_seed_uuid(self, label: str, random_bits: int = 128) -> str:
        """
        Create random UUID. If SOURCE_DATE_EPOCH is present use
        SOURCE_DATE_EPOCH + label name as seed
        """
        sde = os.environ.get('SOURCE_DATE_EPOCH')
        if sde:
            label_seed = reduce(lambda x, y: x + y, map(ord, label))
            epoch_seed = label_seed + int(sde)
            log.info(
                'Using UUID seed SOURCE_DATE_EPOCH:{0} + LABEL:{1}={2}'.format(
                    sde, label, label_seed
                )
            )
            rd = random.Random()
            rd.seed(epoch_seed)
            return format(
                uuid.UUID(int=rd.getrandbits(random_bits))
            )
        else:
            return format(uuid.uuid4())

    def _map_size(self, size: float, from_unit: str, to_unit: str) -> float:
        """
        Return byte size value for given size and unit

        :param float size:
            requested filesystem size. The value is interpreted
            by the given from_unit.

        :param str from_unit: source unit
        :param str to_unit: target unit

        :return: size value in unit: to_unit

        :rtype: float
        """
        unit_map = {
            defaults.UNIT.byte: 1,
            defaults.UNIT.kb: 1024,
            defaults.UNIT.mb: 1048576,
            defaults.UNIT.gb: 1073741824
        }
        byte_size = size * unit_map[from_unit]
        return byte_size / unit_map[to_unit]

    def _fs_size(
        self, size: float, blocksize: int = 1, unit: str = defaults.UNIT.kb
    ) -> str:
        """
        Calculate filesystem size parameter in number of blocksize
        blocks. If the given size is <= 0 the calculation is done from
        the actual size of the block device reduced by the given size

        :param float size:
            requested filesystem size. The value is interpreted
            by the given unit.

        :param int blocksize:
            blocksize as requested from the filesystem creation tool
            for specifying the filesystem size. The value is interpreted
            by the given unit. By default set to: 1

        :param str unit:
            Unit to use for calculations and return value
            Default unit is set to: defaults.UNIT.kb

        :return: an int block count of the specified unit as str

        :rtype: str
        """
        if size > 0:
            result_size = size / blocksize
        else:
            device_name = self.device_provider.get_device()
            device_byte_size = self.device_provider.get_byte_size(device_name)
            requested_byte_size = self._map_size(size, unit, defaults.UNIT.byte)
            result_size = self._map_size(
                (device_byte_size + requested_byte_size) / blocksize,
                from_unit=defaults.UNIT.byte, to_unit=unit
            )
        return format(int(result_size))

    def _apply_attributes(self):
        """
        Apply filesystem attributes
        """
        attribute_map = {
            'synchronous-updates': '+S',
            'no-copy-on-write': '+C'
        }
        for attribute in self.custom_args['fs_attributes']:
            if attribute_map.get(attribute):
                log.info(
                    '--> setting {0} for {1}'.format(
                        attribute, self.filesystem_mount.mountpoint
                    )
                )
                Command.run(
                    [
                        'chattr', attribute_map.get(attribute),
                        self.filesystem_mount.mountpoint
                    ]
                )

    def __exit__(self, exc_type, exc_value, traceback):
        self.umount()
